callJS <- function() {
  # get the parameters from the function that have a value
  message <- Filter(function(x) !is.symbol(x), as.list(parent.frame(1)))
  session <- shiny::getDefaultReactiveDomain()

  # If a linevis widget was passed in, this is during a chain pipeline in the
  # initialization of the widget, so keep track of the desired function call
  # by adding it to a list of functions that should be performed when the widget
  # is ready
  if (methods::is(message$id, "linevis")) {
    widget <- message$id
    message$id <- NULL
    widget$x$api <- c(widget$x$api, list(message))
    return(widget)
  }
  # If an ID was passed, the widget already exists and we can simply call the
  # appropriate JS function
  else if (is.character(message$id)) {
    message$id <- session$ns(message$id)
    method <- paste0("linevis:", message$method)
    session$sendCustomMessage(method, message)
    return(message$id)
  } else {
    stop("The `id` argument must be either a linevis htmlwidget or an ID
         of a linevis htmlwidget.", call. = FALSE)
  }
}


#' Add a new vertical bar at a time point that can be dragged by the user
#'
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param time The date/time to add
#' @param itemId The id of the custom time bar
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   addCustomTime(Sys.Date() - 1, "yesterday")
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Add time bar 24 hours ago")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$btn, {
#'       addCustomTime("graph2d", Sys.Date() - 1, "yesterday")
#'     })
#'   }
#' )
#' }
#' @export
addCustomTime <- function(id, time, itemId) {
  method <- "addCustomTime"
  callJS()
}

#' Adjust the time of a custom time bar
#'
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param time The new date/time
#' @param itemId The id of the custom time bar
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   addCustomTime(Sys.Date(), "yesterday") %>%
#'   setCustomTime(Sys.Date() - 1, "yesterday")
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Set time bar 24 hours ago")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis() %>% addCustomTime(Sys.Date(), "yesterday")
#'     )
#'     observeEvent(input$btn, {
#'       setCustomTime("graph2d", Sys.Date() - 1, "yesterday")
#'     })
#'   }
#' )
#' }
#' @export
setCustomTime <- function(id, time, itemId) {
  method <- "setCustomTime"
  callJS()
}

#' Adjust the time of the current time bar
#'
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param time The new date/time
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   setCurrentTime(Sys.Date())
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Set current time to beginning of today")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$btn, {
#'       setCurrentTime("graph2d", Sys.Date())
#'     })
#'   }
#' )
#' }
#' @export
setCurrentTime <- function(id, time) {
  method <- "setCurrentTime"
  callJS()
}

#' Remove a custom time previously added
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param itemId The id of the custom time bar
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   addCustomTime(Sys.Date() - 1, "yesterday") %>%
#'   addCustomTime(Sys.Date() + 1, "tomorrow") %>%
#'   removeCustomTime("yesterday")
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn0", "Add custom time"),
#'     actionButton("btn", "Remove custom time bar")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$btn0, {
#'       addCustomTime("graph2d", Sys.Date() - 1, "yesterday")
#'     })
#'     observeEvent(input$btn, {
#'       removeCustomTime("graph2d", "yesterday")
#'     })
#'   }
#' )
#' }
#' @export
removeCustomTime <- function(id, itemId) {
  method <- "removeCustomTime"
  callJS()
}

#' Adjust the visible window such that it fits all items
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param options Named list of options controlling the animation. Most common
#' option is \code{"animation" = TRUE/FALSE}. For a full list of options, see
#' the "fit" method in the
#' \href{https://visjs.github.io/vis-timeline/docs/graph2d/#Methods}{official
#' graph2d documentation}
#' @return None, side-effect is Javascript call
#' @examples
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Fit all items")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis(data.frame(
#'         start = c(Sys.Date(), Sys.Date() - 1), content = c("1", "2")
#'       ))
#'     )
#'     observeEvent(input$btn, {
#'       fitWindow("graph2d", list(animation = FALSE))
#'     })
#'   }
#' )
#' }
#' @export
fitWindow <- function(id, options) {
  method <- "fitWindow"
  callJS()
}

#' Move the window such that the given time is centered
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param time The date/time to center around
#' @param options Named list of options controlling the animation. Most common
#' option is \code{"animation" = TRUE/FALSE}. For a full list of options, see
#' the "moveTo" method in the
#' \href{https://visjs.github.io/vis-timeline/docs/graph2d/#Methods}{official
#' graph2d documentation}
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   centerTime(Sys.Date() - 1)
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Center around 24 hours ago")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$btn, {
#'       centerTime("graph2d", Sys.Date() - 1)
#'     })
#'   }
#' )
#' }
#' @export
centerTime <- function(id, time, options) {
  method <- "centerTime"
  callJS()
}

#' Move the window such that given item or items are centered
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param itemId A vector (or single value) of the item ids to center
#' @param options Named list of options controlling mainly the animation.
#' Most common option is \code{"animation" = TRUE/FALSE}. For a full list of
#' options, see the "focus" method in the
#' \href{https://visjs.github.io/vis-timeline/docs/graph2d/#Methods}{official
#' graph2d documentation}
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis(data.frame(
#'           start = c(Sys.Date() - 1, Sys.Date(), Sys.Date() + 1),
#'           content = c("Item 1", "Item 2", "Item 3"))
#' ) %>%
#'   centerItem(1)
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Center around item 1")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis(
#'         data.frame(
#'           start = c(Sys.Date() - 1, Sys.Date(), Sys.Date() + 1),
#'           content = c("Item 1", "Item 2", "Item 3"))
#'       )
#'     )
#'     observeEvent(input$btn, {
#'       centerItem("graph2d", 1)
#'     })
#'   }
#' )
#' }
#' @export
centerItem <- function(id, itemId, options) {
  method <- "centerItem"
  itemId <- as.character(itemId)
  callJS()
}

#' Set the items of a graph2d
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param data A dataframe containing the item data to use.
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis(data.frame(start = Sys.Date(), content = "Today")) %>%
#'   setItems(data.frame(start = Sys.Date() - 1, content = "yesterday"))
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Change the data to yesterday")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis(data.frame(start = Sys.Date(), content = "Today"))
#'     )
#'     observeEvent(input$btn, {
#'       setItems("graph2d",
#'                data.frame(start = Sys.Date() - 1, content = "yesterday"))
#'     })
#'   }
#' )
#' }
#' @export
setItems <- function(id, data) {
  method <- "setItems"
  data <- dataframeToD3(data)
  callJS()
}

#' Set the groups of a graph2d
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param data A dataframe containing the groups data to use.
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis(data = data.frame(
#'   start = c(Sys.Date(), Sys.Date(), Sys.Date() + 1, Sys.Date() + 2),
#'   content = c("one", "two", "three", "four"),
#'   group = c(1, 2, 1, 2)),
#'   groups = data.frame(id = 1:2, content = c("G1", "G2"))
#' ) %>%
#'   setGroups(data.frame(id = 1:2, content = c("Group 1", "Group 2")))
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Change group names")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis(data = data.frame(
#'                start = c(Sys.Date(), Sys.Date(), Sys.Date() + 1, Sys.Date() + 2),
#'                          content = c("one", "two", "three", "four"),
#'                          group = c(1, 2, 1, 2)),
#'               groups = data.frame(id = 1:2, content = c("G1", "G2")))
#'
#'     )
#'     observeEvent(input$btn, {
#'       setGroups("graph2d",
#'                data.frame(id = 1:2, content = c("Group 1", "Group 2")))
#'     })
#'   }
#' )
#' }
#' @export
setGroups <- function(id, data) {
  method <- "setGroups"
  data <- dataframeToD3(data)
  callJS()
}

#' Update the configuration options of a graph2d
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param options A named list containing updated configuration options to use.
#' See the \code{options} parameter of the
#' \code{\link[linevis]{linevis}} function to see more details.
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis(
#'   data.frame(start = Sys.Date(), content = "Today"),
#'   options = list(showCurrentTime = FALSE, orientation = "top")
#' ) %>%
#'   setOptions(list(editable = TRUE, showCurrentTime = TRUE))
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Show current time and allow items to be editable")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis(
#'         data.frame(start = Sys.Date(), content = "Today"),
#'         options = list(showCurrentTime = FALSE, orientation = "top")
#'       )
#'     )
#'     observeEvent(input$btn, {
#'       setOptions("graph2d", list(editable = TRUE, showCurrentTime = TRUE))
#'     })
#'   }
#' )
#' }
#' @export
setOptions <- function(id, options) {
  method <- "setOptions"
  callJS()
}

#' Set the current visible window
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param start The start date/time to show in the graph2d
#' @param end The end date/time to show in the graph2d
#' @param options Named list of options controlling mainly the animation.
#' Most common option is \code{animation = TRUE/FALSE}. For a full list of
#' options, see the "setWindow" method in the
#' \href{https://visjs.github.io/vis-timeline/docs/graph2d/#Methods}{official
#' graph2d documentation}
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   setWindow(Sys.Date() - 1, Sys.Date() + 1)
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     actionButton("btn", "Set window to show between yesterday to tomorrow")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$btn, {
#'       setWindow("graph2d", Sys.Date() - 1, Sys.Date() + 1)
#'     })
#'   }
#' )
#' }
#' @export
setWindow <- function(id, start, end, options) {
  method <- "setWindow"
  callJS()
}

#' Zoom in/out the current visible window
#'
#' @param id graph2d id or a \code{linevis} object (the output from \code{linevis()})
#' @param percent The amount to zoom in or out. Must be a number between 0 and 1.
#' A value of 0.5 means that after zooming out the graph2d will show 50\% more content.
#' @param animation Whether or not to animate the zoom.
#' @return None, side-effect is Javascript call
#' @examples
#' \dontrun{
#' linevis() %>%
#'   zoomIn()
#'
#' linevis() %>%
#'   zoomOut(0.3)
#' }
#'
#' if (interactive()) {
#' library(shiny)
#' shinyApp(
#'   ui = fluidPage(
#'     linevisOutput("graph2d"),
#'     sliderInput("zoom", "Zoom by", min = 0, max = 1, value = 0.5, step = 0.1),
#'     checkboxInput("animate", "Animate?", TRUE),
#'     actionButton("zoomIn", "Zoom IN"),
#'     actionButton("zoomOut", "Zoom OUT")
#'   ),
#'   server = function(input, output) {
#'     output$graph2d <- renderLinevis(
#'       linevis()
#'     )
#'     observeEvent(input$zoomIn, {
#'       zoomIn("graph2d", percent = input$zoom, animation = input$animate)
#'     })
#'     observeEvent(input$zoomOut, {
#'       zoomOut("graph2d", percent = input$zoom, animation = input$animate)
#'     })
#'   }
#' )
#' }
#' @name zoom
NULL

#' @export
#' @rdname zoom
zoomIn <- function(id, percent = 0.5, animation = TRUE) {
  method <- "zoomIn"
  callJS()
}

#' @export
#' @rdname zoom
zoomOut <- function(id, percent = 0.5, animation = TRUE) {
  method <- "zoomOut"
  callJS()
}
